from typing import Annotated

import tyro

from .simple_plots import Bar, Line, Scatter, Smiles

Plot = Annotated[
    Line | Scatter | Bar | Smiles,
    tyro.conf.subcommand(description="Some useful plot cli", name="plot"),
]



import os
from pathlib import Path
from typing import Annotated, Literal

import tyro
from msgspec import Struct, field

from wuxctools.plot import BaseFigSet, HeatFigSet
from wuxctools.vasp.dataread import read_vasp_file

from .parsepro import ProParams


class BandBaseFigSet(BaseFigSet):
    xrange: Annotated[tuple[int, int], tyro.conf.arg(aliases=["-xr"])] = (0, -1)  # pyright: ignore[reportIncompatibleVariableOverride]
    """x index range for figure"""

    yrange: Annotated[tuple[float, float] | None, tyro.conf.arg(aliases=["-yr"])] = (
        field(default_factory=lambda: (-4, 6))
    )

    hide_xticks = True
    ylabel = "Energy (eV)"
    save = ["band.png"]


class BandProParams(ProParams):
    heatparams: Annotated[HeatFigSet, tyro.conf.arg(name="")] = field(
        default_factory=HeatFigSet
    )

    scale: float = 5
    """Specific scatter line scale for pmode 1"""

    fill: Literal["auto", "hollow", "solid"] = "auto"

    """Scatter filling mode:
    - auto: decide based on fc
    - hollow: empty facecolor
    - solid: filled, requires fc or default
    """

    alpha_gamma: float = 0.3
    """use gamma mapping for alpha = (z_norm)^gamma"""


class Band(Struct, tag="band"):
    """
    Plot vasp band figure
    """

    basefigset: BaseFigSet = field(default_factory=BandBaseFigSet)
    pro: BandProParams = field(default_factory=BandProParams)

    files: tyro.conf.Positional[list[Path]] = field(
        default_factory=lambda: [Path("vaspout.h5")]
    )
    """files that vasp read"""

    vaspfileformat: Annotated[
        Literal["h5", "xml", "auto"], tyro.conf.arg(aliases=["-vf"])
    ] = "auto"
    """read vaspfile format [env var: WUXC_VASP_FILE_FORMAT]"""

    efermi: Annotated[float | str | None, tyro.conf.arg(aliases=["-ef"])] = None
    """Fermi energy value, or a path to a file from which the Fermi level is read."""

    spin: Annotated[int | None, tyro.conf.arg(aliases=["-s"])] = None
    """Specify which spin to plot"""

    fix_order: bool = False
    """reorder band for vasp give the wrong order when ncore > 1"""

    labels: Annotated[list[str] | None, tyro.conf.arg(aliases=["-l"])] = None
    """labels for lines"""

    legend: bool = False
    """Whether show legend"""

    export: bool = True
    """Whether export band data"""

    def __post_init__(self):
        from wuxctools.utils import log

        v = os.getenv("WUXC_VASP_FILE_FORMAT", "auto")
        if v == "h5":
            self.vaspfileformat = "h5"
        if v == "xml":
            self.vaspfileformat = "xml"
        elif v == "auto":
            self.vaspfileformat = "auto"
        else:
            log.warning(
                f"Invalid WUXC_VASP_FILE_FORMAT={v!r}, falling back to 'auto'.",
            )
            self.vaspfileformat = "auto"

    def plot_single_file(self, file: Path):
        from .bandplot import BandPlot

        data = read_vasp_file(file, vaspfileformat=self.vaspfileformat)
        BandPlot(self, data)
        self.basefigset()

    def __call__(self):
        import matplotlib.pyplot as plt

        from wuxctools.utils import plot_series, set_style

        set_style(self.basefigset.matplotlibrc)
        plot_series(self.files, self.basefigset.save, self.plot_single_file)

        if self.basefigset.show:
            plt.show()



from functools import cached_property
from pathlib import Path
from typing import cast

import numpy as np
import numpy.typing as npt
from lxml.etree import ElementBase, parse

from .vaspdata import VaspData


class ReadVasprun(VaspData):
    def __init__(
        self, file: str | Path = Path("vasprun.xml"), auto_select_k: bool = False
    ):
        self.file = parse(file)
        if auto_select_k:
            self.k_index = np.where(self.weights == 0)[0]
            if list(self.k_index) == []:
                self.k_index = np.where(self.weights > 0)[0]
        else:
            self.k_index = np.where(self.weights > 0)[0]

    def _get_data_str(self, xpath: str) -> str:
        data = cast(str | None, self.file.xpath(f"string({xpath})"))
        if data is None:
            raise ValueError(f"can't get data by {xpath}")
        else:
            return data

    def _symbols(self) -> list[str]:
        symbolstr = self._get_data_str("/modeling/atominfo/array[@name='atoms']/set[1]")
        return symbolstr.split()[0:-1:2]

    def _ionnum(self) -> int:
        return int(self._get_data_str("/modeling/atominfo/atoms"))

    def _positions(self) -> npt.NDArray[np.floating]:
        position_str = self._get_data_str(
            "/modeling/structure[@name='finalpos']/varray[@name='positions']"
        )
        return np.array(position_str.split(), dtype=float).reshape(-1, 3)

    def _real_cell(self) -> npt.NDArray[np.floating]:
        basis_str = self._get_data_str(
            "/modeling/structure[@name='finalpos']/crystal/varray[@name='basis']"
        )
        return np.array(basis_str.split(), dtype=float).reshape(3, 3)

    # kpoints
    def _kpoints(self) -> npt.NDArray[np.floating]:
        kpoints_str = self._get_data_str("/modeling/kpoints/varray[@name='kpointlist']")
        kpointsarray = np.array(kpoints_str.split(), dtype=float).reshape(-1, 3)
        return kpointsarray[self.k_index]

    def _weights(self) -> npt.NDArray[np.floating]:
        weights_str = self._get_data_str("/modeling/kpoints/varray[@name='weights']")
        return np.array(weights_str.split(), dtype=float)

    def _labels_kpoints(self) -> list[str] | None:
        return None

    def _kpoints_division(self) -> int | None:
        return None

    def _fermi(self) -> float:
        return float(self._get_data_str("/modeling/calculation/dos/i"))

    def _nbands(self) -> int:
        return int(
            self._get_data_str(
                "/modeling/parameters/separator[@name='electronic']/i[@name='NBANDS']"
            )
        )

    def _eigenvalues(self) -> npt.NDArray[np.floating]:
        eigen_str = self._get_data_str("/modeling/calculation/eigenvalues/array/set")
        eigenarray = np.array(eigen_str.split(), dtype=float).reshape(
            -1, len(self.kpoints), self.nbands, 2
        )[:, :, :, 0]
        return eigenarray[:, self.k_index, :]

    def _projected(self) -> npt.NDArray[np.floating]:
        project_str = self._get_data_str("/modeling/calculation/projected/array/set")
        orbitals = cast(
            list[ElementBase],
            self.file.xpath("/modeling/calculation/projected/array/field"),
        )
        orbital_num = len(orbitals)
        projectarray = (
            np.fromstring(project_str, dtype=np.float64, sep=" ")
            .reshape(-1, len(self.kpoints), self.nbands, self.ionnum, orbital_num)
            .transpose([0, 3, 4, 1, 2])
        )
        return projectarray[:, :, :, self.k_index, :]

    # dos

    @cached_property
    def _tdos(self) -> npt.NDArray[np.floating]:
        dos_str = self._get_data_str("/modeling/calculation/dos/total/array/set")
        dosarray = np.array(dos_str.split(), dtype=float).reshape(-1, self.nedos, 3)
        return dosarray

    def _dose(self) -> npt.NDArray[np.floating]:
        return self._tdos[:, :, 0]

    def _dos(self) -> npt.NDArray[np.floating]:
        return self._tdos[:, :, 1]

    def _dosi(self) -> npt.NDArray[np.floating]:
        return self._tdos[:, :, 2]

    def _dospar(self) -> npt.NDArray[np.floating]:
        orbitals = cast(
            list[ElementBase],
            self.file.xpath("/modeling/calculation/dos/partial/array/field"),
        )
        orbital_num = len(orbitals)
        dos_str = self._get_data_str("/modeling/calculation/dos/partial/array/set")
        dosarray = np.array(dos_str.split(), dtype=float).reshape(
            self.ionnum, -1, self.nedos, orbital_num
        )
        return dosarray[:, :, :, 1:].transpose(1, 0, 3, 2)

    def _nedos(self) -> int:
        return int(
            self._get_data_str(
                "/modeling/parameters/separator[@name='dos']/i[@name='NEDOS']"
            )
        )



from contextlib import contextmanager
from contextvars import ContextVar
from pathlib import Path
from typing import Callable, Literal, TypeVar, cast, override

import numpy as np
from matplotlib import colors
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.projections.polar import PolarAxes

CTX_FIG: "ContextVar[Figure | None]" = ContextVar("CTX_FIG", default=None)
CTX_AX: "ContextVar[Axes | list[Axes] | None]" = ContextVar("CTX_AX", default=None)
CTX_AX_INDEX: ContextVar[int] = ContextVar("CTX_AX_INDEX", default=0)


def normalize_axes(ax: "Axes | list[Axes] | np.ndarray | None"):
    import numpy as np
    from matplotlib.axes import Axes

    if isinstance(ax, Axes):
        return ax
    if isinstance(ax, np.ndarray):
        return list(ax.flatten())  # ty:ignore[invalid-argument-type]
    if isinstance(ax, list):
        return ax
    else:
        return None


def get_fig_ax(idx: None | int = None):
    import matplotlib.pyplot as plt

    fig = CTX_FIG.get()
    ax = normalize_axes(CTX_AX.get())

    idx = idx if idx else CTX_AX_INDEX.get()

    if fig is None:
        fig = plt.figure()
        CTX_FIG.set(fig)
    if ax is None:
        ax = fig.add_subplot(1, 1, 1)
        CTX_AX.set(ax)

    if isinstance(ax, list):
        CTX_AX_INDEX.set(idx + 1)
        print(f"now plot on the ax of index {idx % len(ax)}")
        return fig, ax[idx % len(ax)]
    else:
        return fig, ax


@contextmanager
def figure_context(fig: "Figure", ax: "Axes | list[Axes]"):
    token_fig = CTX_FIG.set(fig)
    token_ax = CTX_AX.set(ax)
    token_idx = CTX_AX_INDEX.set(0)
    try:
        yield
    finally:
        CTX_FIG.reset(token_fig)
        CTX_AX.reset(token_ax)
        CTX_AX_INDEX.reset(token_idx)


def set_style(
    rc_file: Path | None = None,
    ncols: int = 1,
    nrows: int = 1,
    use_tex: bool | None = None,
):
    from importlib import resources

    import matplotlib as mpl

    if rc_file is None or not rc_file.exists():
        with resources.path("wuxctools", "wuxc.mplstyle") as rc_path:
            mpl.rc_file(rc_path)
    else:
        mpl.rc_file(rc_file)
    base_size = mpl.rcParams.get("figure.figsize", [3.5433070866, 2.3622047244])
    mpl.rcParams["figure.figsize"] = [
        base_size[0] / 2 * ncols,
        base_size[1] / 2 * nrows,
    ]
    mpl.rcParams["figure.figsize"] = [
        base_size[0] / 2 * ncols,
        base_size[1] / 2 * nrows,
    ]
    if use_tex is not None:
        mpl.rcParams["text.usetex"] = use_tex


T = TypeVar("T", Path, str)


def plot_series(
    single_obj: list[T],
    save: list[str],
    plt_fn: Callable[[T], None],
):
    import matplotlib.pyplot as plt
    from matplotlib.backend_bases import KeyEvent

    fig, ax = get_fig_ax()
    num_figures = len(single_obj)

    if num_figures == 1:
        plt_fn(single_obj[0])
        for savefile in save:
            plt.savefig(savefile)
        return

    if len(save) > 1:
        raise ValueError(
            "only support one save arg which is savedir when plot multiple figures"
        )
    elif len(save) == 1:
        savedir = Path(save[0])
        savedir.mkdir(exist_ok=True, parents=True)
        from tqdm import tqdm

        for file in tqdm(single_obj, desc="Saving figures"):
            save_name = f"{save[0]}/{file}.png"
            plt_fn(file)
            plt.savefig(save_name)
            ax.clear()

    current_index = 0
    path_title = True

    def update_figure(files: list[T], index: int, path_title: bool):
        file = files[index]
        ax.clear()
        plt_fn(file)
        if path_title:
            ax.set_title(f"{file}")

    update_figure(single_obj, 0, path_title)

    fig.canvas.draw()

    def on_key(event: KeyEvent):
        nonlocal current_index, path_title
        if event.key in ["right", "down", "j", "l"]:
            current_index = (current_index + 1) % num_figures
        elif event.key in ["left", "up", "k", "h"]:
            current_index = (current_index - 1) % num_figures
        elif event.key == "t":
            path_title = not path_title
        else:
            return

        update_figure(single_obj, current_index, path_title)
        fig.canvas.draw()

    fig.canvas.mpl_connect("key_press_event", on_key)  # type: ignore


class MyCustomNormalize(colors.Normalize):
    """
    Modified from https://stackoverflow.com/questions/7404116/defining-the-midpoint-of-a-colormap-in-matplotlib
    """

    def __init__(
        self, vmin: float, vmax: float, midpoint: float = 0.0, clip: bool = False
    ):
        self.midpoint: float = midpoint
        colors.Normalize.__init__(self, vmin, vmax, clip)

    @override
    def __call__(self, value: float, clip: bool | None = None):  # type:ignore
        vmin = cast(float, self.vmin)
        vmax = cast(float, self.vmax)
        import numpy as np

        if vmin == vmax:
            return np.full_like(value, 0.5, dtype=np.float64)
        midpoint = self.midpoint

        normalized_min = (
            max(
                0,
                1 / 2 * (1 - abs((midpoint - vmin) / (midpoint - vmax))),
            )
            if (self.midpoint != self.vmax)
            else 0
        )
        normalized_max = (
            min(
                1,
                1 / 2 * (1 + abs((vmax - midpoint) / (midpoint - vmin))),
            )
            if (self.midpoint != self.vmin)
            else 1.0
        )
        normalized_mid = 0.5
        x, y = (
            np.array([self.vmin, self.midpoint, self.vmax]),
            [normalized_min, normalized_mid, normalized_max],
        )
        return np.ma.masked_array(np.interp(value, x, y))


def add_axis(
    ax: Axes,
    direction: Literal["x", "y"] = "y",
    position: float = 1.1,
    ticks: list[float] | None = None,
    ticklabels: list[str] | None = None,
):
    if ax.name != "polar":
        raise ValueError("Axes must be polar")

    polar_ax = cast(PolarAxes, ax)

    if ticks is None:
        data_min, data_max = polar_ax.dataLim.y0, polar_ax.dataLim.y1
        ticks = [data_max, (data_max + data_min) / 2, data_min]

    xmin, xmax = (polar_ax.dataLim.x0, polar_ax.dataLim.x1)
    rmin, rmax = polar_ax.get_rmin(), polar_ax.get_rmax()

    def map_to_symmetric_range(
        y: list[float],
        data_range: tuple[float, float],
        coord_xrange: tuple[float, float],
        labels: list[str] | None = None,
        ndigits: int = 0,
    ) -> tuple[list[float], list[str]]:
        (c_min, c_max), (d_min, d_max) = coord_xrange, data_range
        y_array = np.asanyarray(y)
        mid = (c_min + c_max) / 2
        if d_max == d_min:
            neg = pos = np.full_like(y_array, mid)
        else:
            ratio = (y_array - d_min) / (d_max - d_min)
            pos = mid + ratio * (c_max - mid)
            neg = mid - ratio * (mid - c_min)
        if labels is None:
            labels = [f"{v:.{ndigits}f}" for v in y_array]

        ticks = neg.tolist() + pos[::-1].tolist()
        ticklabels = list(labels) + list(reversed(labels))
        return ticks, ticklabels

    if direction == "x":
        sax = polar_ax.secondary_xaxis(position)

        target_range = (xmin, xmax)
    else:
        sax = polar_ax.secondary_yaxis(position)

        target_range = (rmin, rmax)

    ticks, ticklabels = map_to_symmetric_range(
        ticks, (rmin, rmax), target_range, ticklabels
    )
    sax.set_ticks(ticks, ticklabels)

    return sax



from pathlib import Path

import h5py  # pyright: ignore[reportMissingTypeStubs]
import numpy as np
import numpy.typing as npt

from .vaspdata import VaspData


class ReadVaspout(VaspData):
    def __init__(
        self,
        file: str | Path = "vaspout.h5",
        auto_select_k: bool = False,
        opt: bool | None = None,
    ):
        self.file = h5py.File(file, "r")
        if opt is True or (
            opt is None and self.file.get("input/kpoints_opt") is not None
        ):
            self.prefix1, self.prefix2 = "_opt", "_kpoints_opt"
        else:
            self.prefix1, self.prefix2 = "", ""
        if auto_select_k:
            self.k_index = np.where(self.weights == 0)[0]
            if list(self.k_index) == []:
                self.k_index = np.where(self.weights > 0)[0]
        else:
            self.k_index = np.where(self.weights >= 0)[0]
        # super().__init__()

    # structure

    def _symbols(self) -> list[str]:
        ion_types = np.array(self.file.get("results/positions/ion_types"))
        number_ion_types = np.array(self.file.get("results/positions/number_ion_types"))
        symbols: list[str] = []
        for i, ion in enumerate(ion_types):
            symbols.extend([ion.decode("utf-8")] * number_ion_types[i])
        return symbols

    def _ionnum(self) -> int:
        return int(np.array(self.file.get("results/positions/number_ion_types")).sum())

    def _positions(self) -> npt.NDArray[np.floating]:
        return np.array(self.file.get("results/positions/position_ions"))

    def _real_cell(self) -> npt.NDArray[np.floating]:
        return np.array(self.file.get("results/positions/lattice_vectors"))

    # kpoints
    def _kpoints(self) -> npt.NDArray[np.floating]:
        return np.array(
            self.file.get(f"results/electron_eigenvalues{self.prefix2}/kpoint_coords")
        )

    def _weights(self) -> npt.NDArray[np.floating]:
        return np.array(
            self.file.get(
                f"results/electron_eigenvalues{self.prefix2}/kpoints_symmetry_weight"
            )
        )

    def _labels_kpoints(self) -> list[str] | None:
        labels = self.file.get(f"input/kpoints{self.prefix1}/labels_kpoints")  # pyright: ignore[reportUnknownVariableType]
        if labels is None:
            return None

        labels_array = np.array(labels)
        xticklabels = [i.decode("utf8") for i in labels_array]
        return xticklabels

    def _kpoints_division(self) -> int | None:
        mode = self.file.get(f"input/kpoints{self.prefix1}/mode")  # pyright: ignore[reportUnknownVariableType]
        if np.array(mode) == b"l":
            return int(
                np.array(self.file.get(f"input/kpoints{self.prefix1}/number_kpoints"))
            )
        return None

    def _fermi(self) -> float:
        return float(np.array(self.file.get("results/electron_dos/efermi")))

    def _nbands(self) -> int:
        return self.eigenvalues.shape[-1]

    def _eigenvalues(self) -> npt.NDArray[np.floating]:
        eigenvalues = np.array(
            self.file.get(f"results/electron_eigenvalues{self.prefix2}/eigenvalues")
        )
        return eigenvalues[:, self.k_index, :]

    def _projected(self) -> npt.NDArray[np.floating]:
        return np.array(self.file.get(f"results/projectors{self.prefix2}/par"))[
            :, :, :, self.k_index, :
        ]

    # dos
    def _dos(self) -> npt.NDArray[np.floating]:
        return np.array(self.file.get("results/electron_dos/dos"))

    def _dose(self) -> npt.NDArray[np.floating]:
        return np.array(self.file.get("results/electron_dos/energies"))

    def _dosi(self) -> npt.NDArray[np.floating]:
        return np.array(self.file.get("results/electron_dos/dosi"))

    def _dospar(self) -> npt.NDArray[np.floating]:
        return np.array(self.file.get("results/electron_dos/dospar"))

    def _nedos(self) -> int:
        return int(np.array(self.file.get("results/electron_dos/nedos")))



from pathlib import Path

from .vaspdata import VaspData
from .vaspout import ReadVaspout
from .vasprun import ReadVasprun

__all__ = ["VaspData", "ReadVaspout", "ReadVasprun"]


def read_vasp_file(
    file: Path, vaspfileformat: str = "auto", auto_select_k: bool = False
) -> VaspData:
    if vaspfileformat == "auto":
        if file.suffix in [".h5", ".hdf5"]:
            return ReadVaspout(file, auto_select_k)
        elif file.suffix == ".xml":
            return ReadVasprun(file, auto_select_k)
        else:
            raise ValueError(
                f"Unsupported VASP file format: {file.suffix} for {file.name}"
            )
    elif vaspfileformat in ["h5", "hdf5"]:
        return ReadVaspout(file, auto_select_k)
    elif vaspfileformat == "xml":
        return ReadVasprun(file, auto_select_k)
    else:
        raise ValueError(f"Unsupported VASP file format: {vaspfileformat}")



from abc import ABC, abstractmethod
from functools import cached_property
from pathlib import Path

import numpy as np
import numpy.typing as npt


class VaspData(ABC):
    def __init__(
        self,
        file: str | Path = "vaspout.h5",
        opt: bool | None = None,
        auto_select_k: bool = False,
    ):
        super().__init__()

    @abstractmethod
    def _symbols(self) -> list[str]: ...

    @abstractmethod
    def _ionnum(self) -> int: ...

    @abstractmethod
    def _positions(self) -> npt.NDArray[np.floating]: ...

    @abstractmethod
    def _real_cell(self) -> npt.NDArray[np.floating]: ...

    # kpoints
    @abstractmethod
    def _kpoints(self) -> npt.NDArray[np.floating]: ...

    @abstractmethod
    def _weights(self) -> npt.NDArray[np.floating]: ...

    @abstractmethod
    def _labels_kpoints(self) -> list[str] | None: ...

    @abstractmethod
    def _kpoints_division(self) -> int | None: ...

    @abstractmethod
    def _fermi(self) -> float: ...

    @abstractmethod
    def _nbands(self) -> int: ...

    @abstractmethod
    def _eigenvalues(self) -> npt.NDArray[np.floating]: ...

    @abstractmethod
    def _projected(self) -> npt.NDArray[np.floating]: ...

    @abstractmethod
    def _dos(self) -> npt.NDArray[np.floating]: ...

    @abstractmethod
    def _dose(self) -> npt.NDArray[np.floating]: ...

    @abstractmethod
    def _dosi(self) -> npt.NDArray[np.floating]: ...

    @abstractmethod
    def _dospar(self) -> npt.NDArray[np.floating]: ...

    @abstractmethod
    def _nedos(self) -> int: ...

    @cached_property
    def symbols(self) -> list[str]:
        return self._symbols()

    @cached_property
    def ionnum(self) -> int:
        return self._ionnum()

    @cached_property
    def positions(self) -> npt.NDArray[np.floating]:
        """Atomic positions in Cartesian coordinates.

        Returns
        -------
        numpy.ndarray, shape (ionnum, 3)
            Each row represents the (x, y, z) coordinates of an atom.
        """
        return self._positions()

    @cached_property
    def real_cell(self) -> npt.NDArray[np.floating]:
        """Real-space lattice vectors.

        Returns
        -------
        numpy.ndarray, shape (3, 3)
            Each row represents a lattice vector in Cartesian coordinates.
        """
        return self._real_cell()

    @cached_property
    def rec_cell(self) -> npt.NDArray[np.floating]:
        """Reciprocal lattice vectors.

        Returns
        -------
        numpy.ndarray, shape (3, 3)
            Each row represents a reciprocal lattice vector.
        """
        return np.linalg.inv(self.real_cell).T

    @cached_property
    def kpoints(self) -> npt.NDArray[np.floating]:
        """List of k-points in reciprocal space.

        Returns
        -------
        numpy.ndarray, shape (nkpoints, 3)
            k-point coordinates in reciprocal lattice units.
        """
        return self._kpoints()

    @cached_property
    def weights(self) -> npt.NDArray[np.floating]:
        """Weights of k-points for Brillouin zone integration.

        Returns
        -------
        numpy.ndarray, shape (nkpoints,)
            Weight of each k-point.
        """
        return self._weights()

    @cached_property
    def labels_kpoints(self) -> list[str] | None:
        return self._labels_kpoints()

    @cached_property
    def kpoints_division(self) -> int | None:
        return self._kpoints_division()

    @cached_property
    def fermi(self) -> float:
        return self._fermi()

    # band
    @cached_property
    def nbands(self) -> int:
        return self._nbands()

    @cached_property
    def eigenvalues(self) -> npt.NDArray[np.floating]:
        """A three-dimensional array about eigenvalues

        Returns:
        -------
        <class 'numpy.ndarray'>
            The first dimension represents the different spins
            The second demension represents the different kpoints
            The third demension represents the different bands
        """
        return self._eigenvalues()

    @cached_property
    def projected(self) -> npt.NDArray[np.floating]:
        """A five-dimensional array about projected band

        Returns:
        -------
        <class 'numpy.ndarray'>

            The first dimension represents the different spins
            The second demension represents the different ions
            The third demension represents  different orbitals
            The fourth demension represents the different kpoints
            The fifth demension represents the different bands
        """
        return self._projected()

    @cached_property
    def dose(self) -> npt.NDArray[np.floating]:
        """An 1-dimensional array of energy points for the system.

        Returns
        -------
        <class 'numpy.ndarray'>
            A one-dimensional array representing the energy points.
        """
        return self._dose()

    @cached_property
    def dos(self) -> npt.NDArray[np.floating]:
        """A two-dimensional array of density of states (DOS).

        Returns
        -------
        <class 'numpy.ndarray'>
            The first dimension represents different atoms or spin channels.
            The second dimension represents different energy points.
        """
        return self._dos()

    @cached_property
    def dosi(self) -> npt.NDArray[np.floating]:
        """A two-dimensional array of integrated density of states (IDOS).

        Returns
        -------
        <class 'numpy.ndarray'>
            The first dimension represents different atoms or spin channels.
            The second dimension represents different energy points.
        """
        return self._dosi()

    @cached_property
    def dospar(self) -> npt.NDArray[np.floating]:
        """A four-dimensional array of partial DOS (PDOS).

        Returns
        -------
        <class 'numpy.ndarray'>
            The first dimension represents different spins.
            The second dimension represents different atoms (ions).
            The third dimension represents different orbitals and energy channels.
            The fourth dimension represents the different energy points.
        """
        return self._dospar()

    @cached_property
    def nedos(self) -> int:
        return self._nedos()

    @cached_property
    def orbital_num(self) -> int:
        if "dospar" in self.__dict__:
            return self.dospar.shape[2]
        elif "projected" in self.__dict__:
            return self.projected.shape[2]
        else:
            return self.dospar.shape[2]



import tyro

from .plot.cli import Plot
from .vasp.cli import Vasp

App = Plot | Vasp

app = tyro.cli(
    App,
    config=[
        tyro.conf.OmitSubcommandPrefixes,
        tyro.conf.OmitArgPrefixes,
    ],
)  # ty:ignore[no-matching-overload]



import inspect
import itertools
from functools import cached_property
from pathlib import Path
from typing import cast

import matplotlib as mpl
import numpy as np
import numpy.typing as npt
from matplotlib.artist import Artist
from matplotlib.collections import LineCollection
from matplotlib.lines import Line2D

from wuxctools.utils import HAS_PYMATGEN, get_fig_ax, log

from ..dataread import VaspData, read_vasp_file
from . import Band
from .parsepro import Group


class BandPlot:
    def __init__(self, params: Band, data: VaspData):
        self.params = params
        self.fig, self.ax = get_fig_ax()
        self.data = data

        if isinstance(self.params.efermi, str):
            efermi: float = read_vasp_file(Path(self.params.efermi)).fermi
            self.plot_fermi = efermi
        elif self.params.efermi is not None:
            self.plot_fermi = self.params.efermi
        else:
            self.plot_fermi = self.data.fermi

        self.eigenvalues = (
            np.sort(self.data.eigenvalues, axis=2)
            if self.params.fix_order
            else self.data.eigenvalues
        )

        self.handles: list[Artist] = []

        self.set_style()
        self.gap = self.get_gap()
        if self.params.pro.project == []:
            self.plot_band()
        else:
            self.groups = self.params.pro.parse(
                self.data, self.params.basefigset.colors
            )
            self.plot_proband()

        if self.params.legend:
            self.ax.legend(handles=self.handles, loc="best")

        if self.params.export:
            self.export()
            ...

    @cached_property
    def xlist(self) -> list[float]:
        kpoints_real = [
            np.dot(kpoint, self.data.rec_cell) for kpoint in self.data.kpoints
        ]

        length = (
            np.linalg.norm(kpoint1 - kpoint2)
            for kpoint1, kpoint2 in zip(
                kpoints_real, kpoints_real[0:1] + list(kpoints_real)[:-1]
            )
        )
        return [float(i) * np.pi * 2 for i in itertools.accumulate(length)]

    @cached_property
    def k_tick_labels(self) -> dict[float, str]:
        def symbol_to_latex(symbol: str) -> str:
            if not symbol[0].isalpha() and not (
                symbol.startswith("$") and symbol.endswith("$")
            ):
                return f"${symbol}$"
            return symbol

        def handle_kpoint_labels(klist: list[str]) -> list[str]:
            result: list[str] = []
            result.append(symbol_to_latex(klist[0]))
            for k1, k2 in zip(klist[2::2], klist[1:-1:2]):
                k1_new, k2_new = symbol_to_latex(k1), symbol_to_latex(k2)
                if k1_new == k2_new:
                    result.append(k1_new)
                else:
                    result.append(f"{k2_new}/{k1_new}")
            result.append(klist[-1])
            return result

        def get_k_labels(
            ks: list[np.ndarray],
            labels_dict: dict[str, tuple[float, float, float]] = {},
            default_labels_dict: dict[str, tuple[float, float, float]] = {
                r"\Gamma": (0.0, 0.0, 0.0),
                "K": (1 / 3, 1 / 3, 0.0),
                "M": (0.5, 0.0, 0.0),
            },
        ):
            """Assign labels to K-points based on nearest distance."""
            result: list[str] = []
            if labels_dict == {}:
                if HAS_PYMATGEN:
                    log.info(
                        "No label dictionary provided. Attempting to infer k-point labels using pymatgen..."
                    )
                    from pymatgen.core import (  # pyright: ignore[reportMissingTypeStubs]
                        Lattice,
                        Structure,
                    )
                    from pymatgen.symmetry.bandstructure import (  # pyright: ignore[reportMissingTypeStubs]
                        HighSymmKpath,  # pyright: ignore[reportMissingTypeStubs]
                    )

                    lattice = Lattice(self.data.real_cell)

                    structure = Structure(
                        lattice, self.data.symbols, self.data.positions
                    )

                    try:
                        kpath = HighSymmKpath(structure).kpath
                        raw_dict: dict[str, np.ndarray] = kpath["kpoints"]  # pyright: ignore[reportUnknownVariableType, reportOptionalSubscript]
                        labels_dict = {
                            label: (
                                float(coords[0]),  # pyright: ignore[reportUnknownArgumentType]
                                float(coords[1]),  # pyright: ignore[reportUnknownArgumentType]
                                float(coords[2]),  # pyright: ignore[reportUnknownArgumentType]
                            )
                            for label, coords in raw_dict.items()  # pyright: ignore[reportUnknownVariableType]
                        }
                    except Exception:
                        log.warning(
                            "Pymatgen failed to determine k-point labels. Falling back to default labels."
                        )
                        labels_dict = default_labels_dict
                else:
                    log.warning("No label dictionary provided. Using  default  labels.")
                    labels_dict = default_labels_dict

            log.warning(
                f"Assigning labels based on k-point distance:\n {labels_dict} \n Results may be inaccurate."
            )

            for k in ks:
                nearest_label: str = min(
                    labels_dict.keys(),
                    key=lambda label: float(
                        np.linalg.norm(k - np.asarray(labels_dict[label]))
                    ),
                )
                result.append(symbol_to_latex(nearest_label))
            return result

        k_rec = list(self.data.kpoints[:: self.data.kpoints_division])
        k_rec.append(self.data.kpoints[-1])

        k_x = ([0.0] + self.xlist)[:: self.data.kpoints_division]
        k_labels = (
            get_k_labels(k_rec)
            if self.data.labels_kpoints is None
            else handle_kpoint_labels(self.data.labels_kpoints)
        )

        return dict(zip(k_x, k_labels))

    @cached_property
    def ylist(self) -> npt.NDArray[np.floating]:
        return self.eigenvalues - self.plot_fermi

    def get_gap(
        self,
        vbms: list[int] | None = None,
    ) -> list[float]:
        def is_metal_or_vbm(ylist: np.ndarray) -> bool | int:
            nbands = ylist.shape[1]
            extremum = np.array(
                [[ylist[:, n].min(), ylist[:, n].max()] for n in range(nbands)]
            )

            if (extremum[:, 0] * extremum[:, 1]).min() > 0:
                maxs = extremum[:, 1]
                vbm = int(np.count_nonzero(abs(maxs) - maxs))
                return vbm
            else:
                return True

        def get_kpoints_vbm(ylist: np.ndarray, kpoints: np.ndarray):
            index = np.where(ylist == ylist.max())
            return kpoints[index]

        def get_kpoints_cbm(ylist: np.ndarray, kpoints: np.ndarray):
            index = np.where(ylist == ylist.min())
            return kpoints[index]

        gaps: list[float] = []
        log_str = (
            inspect.cleandoc(
                f"""
            Fermi level used for plotting is {self.plot_fermi:.6f} eV. (read from specified file: {Path(self.params.efermi)}).
            Fermi level in the plot file is {self.data.fermi:.6f} eV.
            """
            )
            if isinstance(self.params.efermi, str)
            else f"Fermi level used for plotting is {self.plot_fermi:.6f} eV\n"
        )

        for i, spin in enumerate(self.ylist):
            vbm = is_metal_or_vbm(spin) if vbms is None else vbms[i]
            log_str += f"[yellow]{'=' * 30} {i + 1}th spin {'=' * 30} [/yellow]\n"
            if vbm is True:
                gaps.append(0.0)
                log_str += "Fermi level intersects energy bands, [blue]Metal[/blue]\n"
            else:
                e_vbm = float(spin[:, vbm - 1].max())
                k_vbm = get_kpoints_vbm(spin[:, vbm - 1], self.data.kpoints)
                vbm_energy = spin[:, vbm - 1].max() + self.data.fermi
                cbm_energy = spin[:, vbm].min() + self.data.fermi
                e_cbm = float(spin[:, vbm].min())
                k_cbm = get_kpoints_cbm(spin[:, vbm], self.data.kpoints)
                gap = e_cbm - e_vbm
                if gap <= 0:
                    gaps.append(0.0)
                    log_str += (
                        "Fermi level intersects energy bands, [blue]Metal[/blue]\n"
                    )

                else:
                    gaps.append(gap)
                    log_str += f"vbm locates at{k_vbm} of [red]{vbm}th[/red] band, vbm energy is [orange1]{vbm_energy:.6f} eV[/orange1]\n"

                    log_str += f"cbm locates at {k_cbm} of [red]{vbm + 1}th[/red] band, cbm energy is [orange1]{cbm_energy:.6f}[/orange1] eV\n"
                    log_str += f"band gap is {gap}\n"
        log.info(log_str)

        return gaps

    def set_style(self):
        xindex_start, xindex_end = cast(tuple[int, int], self.params.basefigset.xrange)
        print(xindex_start, xindex_end)

        self.params.basefigset.xrange = (
            self.xlist[xindex_start],
            self.xlist[xindex_end],
        )
        k_ticks = sorted(self.k_tick_labels.keys())
        self.params.basefigset.xticks = (
            k_ticks
            if self.params.basefigset.xticks is None
            else self.params.basefigset.xticks
        )
        self.params.basefigset.xticklabels = (
            [self.k_tick_labels[k] for k in k_ticks]
            if self.params.basefigset.xticklabels is None
            else self.params.basefigset.xticklabels
        )

        y_major_tick_size = mpl.rcParams["ytick.major.size"]
        self.ax.axhline(
            0,
            ls=(0, (y_major_tick_size, y_major_tick_size)),
            c="black",
            lw=mpl.rcParams["ytick.major.width"],
            zorder=0,
        )

        for i in k_ticks:
            self.ax.axvline(
                i, c="lightgrey", zorder=1, lw=mpl.rcParams["ytick.major.width"]
            )

    @cached_property
    def lc_list(self) -> list[LineCollection]:
        x = np.array(self.xlist)  # (K,)
        x_mid = (x[:-1] + x[1:]) / 2  # (K-1,)
        x_left, x_right = np.r_[x[0], x_mid], np.r_[x_mid, x[-1]]  # (K,)
        x_left2 = np.broadcast_to(x_left[None, :, None], self.ylist.shape)  # (S,K,B)
        x_right2 = np.broadcast_to(x_right[None, :, None], self.ylist.shape)  # (S,K,B)
        y_mid = (self.ylist[:, :-1, :] + self.ylist[:, 1:, :]) / 2  # (S, K-1, B)
        y_left = np.concatenate([self.ylist[:, 0:1, :], y_mid], axis=1)  # (S, K, B)
        y_right = np.concatenate([y_mid, self.ylist[:, -1:, :]], axis=1)  # (S, K, B)

        segments_list = [
            np.stack(
                [
                    np.stack([x_left2[i], x_right2[i]], axis=-1),  # (S,K,B,2)
                    np.stack([y_left[i], y_right[i]], axis=-1),  # (S,K,B,2)
                ],
                axis=-1,
            ).reshape(-1, 2, 2)
            for i in range(self.ylist.shape[0])
        ]
        lc_list = [
            LineCollection(list(segments), rasterized=True)
            for segments in segments_list
        ]

        for lc in lc_list:
            lc.set_capstyle("round")

        return lc_list

    def plot_band(self):
        labels = ["↑", "↓"] if self.params.labels is None else self.params.labels
        lc_list = (
            [self.lc_list[self.params.spin]]
            if self.params.spin is not None
            else self.lc_list
        )
        for lc, color, label in zip(lc_list, self.params.basefigset.colors, labels):
            self.ax.add_collection(lc)
            lc.set_color(color)
            lc.set_zorder(2)
            self.handles.append(Line2D([], [], c=color, label=label))

    def export(self):
        filenames = ["band_up.txt", "band_down.txt"]
        for i, eigen in enumerate(self.eigenvalues):
            self.export_band(filenames[i], eigen)
        # self.save_pro_band()

    def export_band(
        self,
        filename: str,
        eigen: npt.NDArray[np.floating],
        weight: npt.NDArray[np.floating] | None = None,
    ):
        with open(filename, "w") as f:
            for i in range(eigen.shape[1]):  # 遍历每条能带
                if weight is None:
                    data = np.column_stack((self.xlist, eigen[:, i]))
                else:
                    data = np.column_stack((self.xlist, eigen[:, i], weight[:, i]))

                np.savetxt(f, data, fmt="%.10f")
                f.write("\n")

    def save_pro_band(self):
        proarray_list = [self.get_proarray(group) for group in self.groups]
        spins: list[int] = []
        if self.params.spin is None:
            if len(self.eigenvalues) == 1:
                spins = [0]
            elif len(self.eigenvalues) == 2:
                spins = [0, 1]
            elif len(self.eigenvalues) == 4:
                spins[0]
        else:
            spin = [self.params.spin]

        for spin in spins:
            eigen = self.eigenvalues[int(spin)]

            for i, pro in enumerate(proarray_list):
                weight = pro[spin]
                self.export_band(f"fatband_spin{spin}_{i}.txt", eigen, weight)

    def get_proarray(self, group: Group) -> np.ndarray:
        if not group:
            raise ValueError

        proarray = np.add.reduce(
            [
                self.data.projected[:, atoms][:, :, orbitals].sum(axis=(1, 2))
                for atoms, orbitals in group
            ]
        )
        if self.params.spin is not None:
            proarray = proarray[self.params.spin : self.params.spin + 1]
        elif len(self.data.projected) == 2 and self.params.pro.pmode == 1:
            proarray[1] = -proarray[1]
        elif len(self.data.projected) == 2 and self.params.pro.pmode != 1:
            ...
        else:
            proarray = proarray[0:1, :, :]
        return proarray

    def plot_proband(self):
        if self.params.pro.pmode == 1:
            pass
            self.pband1()
        elif self.params.pro.pmode == 2:
            self.pband2()

    def pband1(self):
        proarray = self.get_proarray(self.groups[0])
        lc_array = proarray.reshape(-1)
        if proarray.shape[0] == 2:
            segments_list = [seg for lc in self.lc_list for seg in lc.get_segments()]
            lc = LineCollection(segments_list)
        else:
            lc = self.lc_list[0]

        lc.set_capstyle("round")
        lc.set_array(lc_array)
        self.ax.add_collection(lc)
        self.params.pro.heatparams(lc)

    def pband2(self):
        colors = iter(self.params.basefigset.colors)
        xlist = np.broadcast_to(np.array(self.xlist)[None, :, None], self.ylist.shape)

        pro_max = max(self.get_proarray(group).max() for group in self.groups)

        labels = (
            self.params.labels
            if self.params.labels
            else [group.label for group in self.groups]
        )

        for group, label in zip(self.groups, labels):
            proarray = self.get_proarray(group)
            color = next(colors)
            for x, y, z in zip(xlist, self.ylist, proarray):
                z_norm = z / pro_max
                alpha = np.power(z_norm, self.params.pro.alpha_gamma)

                if self.params.pro.fill == "hollow":
                    fc = "none"
                    ec = color
                    alpha = 1

                elif self.params.pro.fill == "solid":
                    fc = color
                    ec = "none"
                else:
                    fc = color
                    ec = fc
                self.ax.scatter(
                    x,
                    y,
                    s=z_norm * self.params.pro.scale,
                    fc=fc,
                    ec=ec,
                    alpha=alpha,
                    label=label,
                )

                self.handles.append(
                    Line2D([], [], c=color, marker="o", ls="", label=label)
                )



from functools import cached_property
from pathlib import Path
from typing import Any, Callable

import matplotlib as mpl
import numpy as np
from matplotlib import colors
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from matplotlib.patches import Polygon
from matplotlib.typing import ColorType

from wuxctools.utils import get_fig_ax

from ..dataread import read_vasp_file
from ..dataread.vaspdata import VaspData
from . import Dos


def fillplot(x: np.ndarray, y: np.ndarray, ax: Axes, **kwargs: Any):
    ax.fill(x, y, **kwargs)


def lineplot(x: np.ndarray, y: np.ndarray, ax: Axes, **kwargs: Any):
    ax.plot(x, y, **kwargs)


# copy from https://stackoverflow.com/questions/29321835/is-it-possible-to-get-color-gradients-under-curve-in-matplotlib
def gradient_fill(
    x: np.ndarray,
    y: np.ndarray,
    ax: Axes,
    fill_color: None | ColorType = None,
    **kwargs: Any,
):
    """
    Plot a line with a linear alpha gradient filled beneath it.

    Parameters
    ----------
    x, y : array-like
        The data values of the line.
    fill_color : a matplotlib color specifier (string, tuple) or None
        The color for the fill. If None, the color of the line will be used.
    ax : a matplotlib Axes instance
        The axes to plot on. If None, the current pyplot axes will be used.
    Additional arguments are passed on to matplotlib's ``plot`` function.

    """

    (line,) = ax.plot(x, y, **kwargs)
    if fill_color is None:
        fill_color = line.get_color()

    zorder = line.get_zorder()
    alpha = line.get_alpha()
    alpha = 1.0 if alpha is None else alpha

    z = np.empty((100, 1, 4), dtype=float)
    rgb = colors.colorConverter.to_rgb(fill_color)
    z[:, :, :3] = rgb
    z[:, :, -1] = np.linspace(0, alpha, 100)[:, None]
    xmin, xmax, ymin, ymax = x.min(), x.max(), y.min(), y.max()
    if ymin < 0:
        ymin, ymax = ymax, ymin
    im = ax.imshow(
        z, aspect="auto", extent=(xmin, xmax, ymin, ymax), origin="lower", zorder=zorder
    )

    xy = np.column_stack([x, y])
    xy = np.vstack([[xmin, ymin], xy, [xmax, ymin], [xmin, ymin]])
    clip_path = Polygon(xy, facecolor="none", edgecolor="none", closed=True)
    ax.add_patch(clip_path)
    im.set_clip_path(clip_path)

    ax.autoscale(True)


class DosPlot:
    def __init__(self, params: Dos, data: VaspData):
        self.params = params
        self.fig, self.ax = get_fig_ax()

        self.data = data

        if isinstance(self.params.efermi, str):
            efermi: float = read_vasp_file(Path(self.params.efermi)).fermi
            self.plot_fermi = efermi
        elif self.params.efermi is not None:
            self.plot_fermi = self.params.efermi
        else:
            self.plot_fermi = self.data.fermi

        plot_fn_dict = {"grad": gradient_fill, "fill": fillplot, "line": lineplot}
        self.plot_fn = plot_fn_dict[self.params.style]
        self.handles: list[Artist] = []

        if self.params.tdos:
            self._plotdos(
                self.total_dos, plot_fn_dict[self.params.tdos_style], color="black"
            )
            self.handles.append(Line2D([], [], c="black", label=self.params.tlabel))
        self.plot_pdos()
        if self.params.legend:
            self.ax.legend(handles=self.handles, loc="best")
        self.set_style()

    @cached_property
    def xlist(self) -> np.ndarray:
        return self.data.dose - self.plot_fermi

    @cached_property
    def total_dos(self) -> np.ndarray:
        return self.data.dos

    def _plotdos(self, dos: np.ndarray, plot_fn: Callable[..., None], **kwargs: Any):
        if self.params.spin is not None:
            plot_fn(self.xlist, dos[self.params.spin], self.ax, **kwargs)
        elif len(self.total_dos) == 2:
            plot_fn(self.xlist, dos[0], self.ax, **kwargs)
            plot_fn(self.xlist, -dos[1], self.ax, **kwargs)
        else:
            plot_fn(self.xlist, dos[0], self.ax, **kwargs)

    def plot_pdos(self):
        groups = self.params.pro.parse(self.data, self.params.basefigset.colors)
        for group, color in zip(groups, self.params.basefigset.colors):
            pdos = np.add.reduce(
                [
                    self.data.dospar[:, atoms][:, :, orbitals].sum(axis=(1, 2))
                    for atoms, orbitals in group
                ]
            )
            self._plotdos(pdos, self.plot_fn, c=color)
            self.handles.append(Line2D([], [], c=color, label=group.label))

    def set_style(self):
        self.params.basefigset.xrange = (
            (min(self.xlist), max(self.xlist))
            if self.params.basefigset.xrange is None
            else self.params.basefigset.xrange
        )

        self.ax.axhline(
            0,
            c="black",
            lw=mpl.rcParams["ytick.major.width"],
            zorder=0,
        )



from rich.console import Console
from rich.panel import Panel

from .plot import figure_context, get_fig_ax, plot_series, set_style

__all__ = ["get_fig_ax", "plot_series", "set_style", "figure_context"]

console = Console()


class SimplePanelLogger:
    def __init__(self, name: str):
        self.name = name

    def log(self, level: str, message: str):
        panel = Panel(
            message,
            title=f"{self.name}: {level}",
            border_style={"INFO": "blue", "WARNING": "yellow", "ERROR": "red"}.get(
                level, "white"
            ),
            highlight=True,
        )
        console.print(panel)

    def info(self, msg: str):
        self.log("INFO", msg)

    def warning(self, msg: str):
        self.log("WARNING", msg)

    def error(self, msg: str):
        self.log("ERROR", msg)


# 使用
log = SimplePanelLogger("wuxctools")


def _try_import(module_name: str):
    try:
        module = __import__(module_name)
        return module, True
    except ImportError:
        return None, False


pymatgen, HAS_PYMATGEN = _try_import("pymatgen")



from pathlib import Path
from typing import Annotated, Literal

import tyro
from msgspec import Struct, field

from wuxctools.plot.common_params import BaseFigSet
from wuxctools.vasp.dataread import read_vasp_file
from wuxctools.vasp.subcommand.parsepro import ProParams


class DosBaseFigSet(BaseFigSet):
    xlabel = "Energy (eV)"
    ylabel = "DOS (states/eV)"
    save = ["dos.png"]


class Dos(Struct, tag="dos"):
    basefigset: DosBaseFigSet = field(default_factory=DosBaseFigSet)
    pro: ProParams = field(default_factory=ProParams)

    files: tyro.conf.Positional[list[Path]] = field(
        default_factory=lambda: [Path("vaspout.h5")]
    )
    """files that vasp read"""

    vaspfileformat: Annotated[
        Literal["h5", "xml", "auto"], tyro.conf.arg(aliases=["-vf"])
    ] = "auto"
    """read vaspfile format [env var: WUXC_VASP_FILE_FORMAT]"""

    efermi: Annotated[float | str | None, tyro.conf.arg(aliases=["-ef"])] = None
    """Fermi energy value, or a path to a file from which the Fermi level is read."""

    spin: Annotated[int | None, tyro.conf.arg(aliases=["-s"])] = None
    """Specify which spin to plot"""

    style: Literal["line", "fill", "grad"] = "grad"
    "Plot style"

    tdos: bool = True
    "Whether plot tdos"
    tlabel: str = "TDOS"
    "plot tdos label"

    tdos_style: Literal["line", "fill", "grad"] = "line"
    "total dos plot style"

    tdos_color: str = "black"
    "total dos plot color"

    legend: bool = False
    """Whether show legend"""

    def plot_single_file(self, file: Path):
        from .dosplot import DosPlot

        data = read_vasp_file(file)

        DosPlot(self, data)

        self.basefigset()
        pass

    def __call__(self) -> None:
        import matplotlib.pyplot as plt

        from wuxctools.utils import plot_series, set_style

        set_style(self.basefigset.matplotlibrc)
        plot_series(self.files, self.basefigset.save, self.plot_single_file)

        if self.basefigset.show:
            plt.show()



import itertools
from typing import TYPE_CHECKING, Annotated, Literal

import tyro
from msgspec import Struct, field

if TYPE_CHECKING:
    from ..dataread import VaspData

ORBITALS_STR_ALL = (
    "s py pz px dxy dyz dz2 dxz dx2-y2 fy3x2 fxyz fyz2 fz3 fxz2 fzx2 fx3".split()
)

ORBITAL_GROUPS = {
    "p": ["px", "py", "pz"],
    "d": ["dxy", "dyz", "dz2", "dxz", "dx2-y2"],
    "f": ["fy3x2", "fxyz", "fyz2", "fz3", "fxz2", "fzx2", "fx3"],
}

ORBITAL_NUMBERS = {
    "p": ["p1", "p2", "p3"],
    "d": ["d4", "d5", "d6", "d7", "d8"],
    "f": ["f9", "f10", "f11", "f12", "f13", "f14", "f15"],
}


class Group(Struct):
    atoms_dict: dict[str, list[int]]
    orbitals_dict: dict[str, list[int]]
    colors: list[str]
    index: int = 0
    pairs: set[tuple[int, int]] = field(default_factory=set)
    atoms: list[list[int]] = field(default_factory=list)
    orbitals: list[list[int]] = field(default_factory=list)
    atoms_label_list: list[str] = field(default_factory=list)
    orbitals_label_list: list[str] = field(default_factory=list)
    label: str = ""

    def __iter__(self):
        return iter((list(zip(self.atoms, self.orbitals))))

    def __len__(self):
        return len(self.atoms)

    def _check_unique(self, name: str, lst: list[int]):
        import sys

        from wuxctools.utils import log

        if len(lst) != len(set(lst)):
            log.error(
                f"[red]{lst}[/red] expands to {lst} which contains duplicate {name} indices."
            )
            sys.exit(1)

    def add(self, atoms: list[int], orbitals: list[int]):
        self._check_unique("atom", atoms)
        self._check_unique("orbital", orbitals)
        self._get_atoms_label(atoms, orbitals)
        self.atoms.append(atoms)
        self.orbitals.append(orbitals)

    def check(self):
        import sys

        from wuxctools.utils import log

        for atoms_list, orbitals_list in zip(self.atoms, self.orbitals):
            for a, o in itertools.product(atoms_list, orbitals_list):
                pair = (a, o)
                if pair in self.pairs:
                    log.error(
                        f"Duplicate projection pair detected: atom={a}, orbital={o}. "
                        "Please check your projection rules."
                    )
                    sys.exit(1)
                self.pairs.add(pair)
        log_str = f"projected atoms and orbitals of your choice {self.index} is\n"

        label_list: list[str] = []
        for a, o in zip(self.atoms_label_list, self.orbitals_label_list):
            log_str += f"orbitals [yellow]{o}[/yellow] for atoms [yellow]{a}[/yellow]\n"
            label_list.append(f"{a}:{o}")
        self.label = "+".join(label_list)

        log.info(log_str)

    def _coverage_label(
        self, covered_indices: list[int], index_dict: dict[str, list[int]]
    ) -> str:
        result: list[str] = []

        for elem, idxs in index_dict.items():
            if not elem[0].isalpha():
                continue
            covered = sorted(i for i in idxs if i in covered_indices)
            if not covered:
                continue

            if len(covered) == len(idxs):
                result.append(elem)
            else:
                result.extend(list(f"{elem}{i}" for i in covered))

        cleaned = result[:]

        for group, subs in ORBITAL_GROUPS.items():
            has_group = group in result

            if has_group:
                cleaned = [x for x in cleaned if x not in subs]
            else:
                cleaned = [x for x in cleaned if x not in ORBITAL_NUMBERS[group]]
        return "+".join(cleaned)

    def _get_atoms_label(self, atoms: list[int], orbitals: list[int]):
        atoms_label = self._coverage_label(atoms, self.atoms_dict)
        orbitals_label = self._coverage_label(orbitals, self.orbitals_dict)
        self.atoms_label_list.append(atoms_label)
        self.orbitals_label_list.append(orbitals_label)


class ProParams(Struct):
    pmode: Literal[1, 2, 3] = 1

    project: Annotated[
        tyro.conf.UseAppendAction[list[list[str]]], tyro.conf.arg(aliases=["-p"])
    ] = field(default_factory=lambda: [])
    """
    Add a projection rule. Can be used multiple times.
    ATOMS:
      Atom selectors (comma-separated):
        SYMBOL        e.g. Mn, Bi
        SYMBOL+INDEX  e.g. Bi3   (3rd Bi)
        INDEX         e.g. 0, 1, 12
    ORBITS:
      Orbital selectors (comma-separated):
        Names: s, p, d, f, px, py, pz, dxy, dxz, dyz, dx2-y2, dz2
        Or indices: 0,1,2,3... (code-dependent ordering)
    Examples:
      -p Mn:p,d            (Mn: p + d)
      -p Bi3:p             (Bi3: p)
      -p Bi:1,2,3          (Bi atom #1,2,3: orbital index 1/2/3)
      -p 0,1:dxy           (atom 0 & 1: dxy)            
    """

    def parse(self, data: "VaspData", colors: list[str]) -> list[Group]:
        import sys

        from wuxctools.utils import log

        atoms_dict: dict[str, list[int]] = {}
        for i, symbol in enumerate(data.symbols):
            atoms_dict.setdefault(symbol, []).append(i)
            atoms_dict.setdefault(str(i), []).append(i)

        orbitals_dict: dict[str, list[int]] = {}
        for i, orbital in enumerate(ORBITALS_STR_ALL[0 : data.orbital_num]):
            orbitals_dict.setdefault(orbital, []).append(i)
            orbitals_dict.setdefault(str(i), []).append(i)
            if "p" in orbital:
                orbitals_dict.setdefault("p", []).append(i)
            if "d" in orbital:
                orbitals_dict.setdefault("d", []).append(i)
            if "f" in orbital:
                orbitals_dict.setdefault("f", []).append(i)

        def lookup(d: dict[str, list[int]], key: str, kind: str) -> list[int]:
            try:
                return d[key]
            except KeyError:
                log.error(
                    f"Invalid {kind!s} selector: {key!r}, you can only choose {kind!s} in {sorted(d.keys())}"
                )
                sys.exit(1)

        result: list[Group] = []
        for index, subpro in enumerate(self.project):
            group = Group(atoms_dict, orbitals_dict, colors, index)
            for subgroup in subpro:
                atoms, orbitals = subgroup.split(":")

                atoms_list = [
                    idx
                    for sel in atoms.split(",")
                    for idx in lookup(atoms_dict, sel, "atom")
                ]

                orbitals_list = [
                    idx
                    for sel in orbitals.split(",")
                    for idx in lookup(orbitals_dict, sel, "orbital")
                ]
                group.add(atoms_list, orbitals_list)
            group.check()

            result.append(group)
        return result



from pathlib import Path

import tyro
from msgspec import Struct, field

from wuxctools.plot import BaseFigSet


class Smiles(Struct, tag="Smiles"):
    """
    Plot Smiles figure
    """

    smiles_str: tyro.conf.Positional[list[str] | None] = None
    file: Path | None = None
    unpack: bool = True
    cols: int = 1
    width: int = 800
    height: int = 800
    basefigset: BaseFigSet = field(default_factory=BaseFigSet)

    def plot_sigle_smiles(self, smiles_str: str) -> None:
        from rdkit import Chem
        from rdkit.Chem import Draw

        from wuxctools.utils import get_fig_ax, log

        log.info(f"Current Plot smiles is [yellow]{smiles_str}[/yellow]")
        mol = Chem.MolFromSmiles(smiles_str)
        if mol is None:  # pyright: ignore[reportUnnecessaryComparison]
            return
        _, ax = get_fig_ax()

        img = Draw.MolToImage(mol, size=(1000, 1000))  # pyright: ignore[reportUnknownVariableType]
        ax.imshow(img)  # pyright: ignore[reportUnknownArgumentType]
        ax.axis("off")

        self.basefigset()

    def __call__(self) -> None:
        import matplotlib.pyplot as plt

        from wuxctools.utils import plot_series, set_style

        set_style(self.basefigset.matplotlibrc, use_tex=False)
        if self.smiles_str:
            plot_series(self.smiles_str, self.basefigset.save, self.plot_sigle_smiles)
        if self.file:
            import polars as pl

            df = pl.read_csv(self.file)
            smiles_list: list[str] = df[:, 0].to_list()

            plot_series(smiles_list, self.basefigset.save, self.plot_sigle_smiles)

        if self.basefigset.show:
            plt.show()



import os
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Literal

import tyro
from msgspec import Struct, field

if TYPE_CHECKING:
    from matplotlib.cm import ScalarMappable


class BaseFigSet(Struct):
    from_cli: Annotated[bool, tyro.conf.Suppress] = False

    matplotlibrc: Annotated[Path, tyro.conf.arg()] = field(
        default_factory=lambda: Path(
            os.getenv("WUXC_MATPLOTLIBRC_FILE", "~/.config/wuxctools/matplotlibrc")
        )
    )
    """ matplotlibrc style file [env var: WUXC_MATPLOTLIBRC_FILE] """

    figsize: tuple[float, float] | None = None
    """size for figure"""

    xrange: Annotated[tuple[float, float] | None, tyro.conf.arg(aliases=["-xr"])] = None
    """x range for figure"""

    yrange: Annotated[tuple[float, float] | None, tyro.conf.arg(aliases=["-yr"])] = None
    """y range for figure"""

    xticks: Annotated[list[float] | None, tyro.conf.arg(aliases=["-xt"])] = None
    """xtick list for Figure"""

    yticks: Annotated[list[float] | None, tyro.conf.arg(aliases=["-yt"])] = None
    """ytick list for Figure"""

    xticklabels: Annotated[list[str] | None, tyro.conf.arg(aliases=["-xtl"])] = None
    """xticklabel list for Figure"""

    yticklabels: Annotated[list[str] | None, tyro.conf.arg(aliases=["-ytl"])] = None
    """yticklabel list for Figure"""

    xlabel: Annotated[str, tyro.conf.arg(aliases=["-xl"])] = ""
    """xlabel for Figure"""

    ylabel: Annotated[str, tyro.conf.arg(aliases=["-yl"])] = ""
    """ylabel for Figure"""

    hide_xticks: bool = False
    """Whether hide xticks"""

    hide_yticks: bool = False
    """Whether hide yticks"""

    n_minor: int = 2
    "Number of minor ticks between two major ticks (AutoMinorLocator n)"

    colors: Annotated[list[str], tyro.conf.arg(aliases=["-c"])] = field(
        default_factory=lambda: [
            "red",
            "blue",
            "orange",
            "green",
            "pink",
            "brown",
        ]
    )
    """colors list for Figure."""

    title: Annotated[str, tyro.conf.arg(aliases=["-t"])] = ""
    """fitle for Figure."""

    show: bool = True
    """Whether show figure"""

    save: list[str] = field(default_factory=lambda: [])
    """Save figure names"""

    def __call__(self):
        from matplotlib.ticker import AutoLocator, AutoMinorLocator

        from wuxctools.utils import get_fig_ax, set_style

        set_style(self.matplotlibrc)

        fig, ax = get_fig_ax()
        if self.figsize:
            fig.set_size_inches(*self.figsize)

        ## title
        ax.set_title(self.title)

        ax.set_xlabel(self.xlabel)
        ax.set_xticks(
            self.xticks
        ) if self.xticks is not None else ax.xaxis.set_major_locator(AutoLocator())
        ax.set_xticklabels(self.xticklabels) if self.xticklabels is not None else ...
        ax.xaxis.set_minor_locator(AutoMinorLocator(self.n_minor))

        ax.set_ylabel(self.ylabel)
        ax.set_yticks(
            self.yticks
        ) if self.yticks is not None else ax.yaxis.set_major_locator(AutoLocator())
        ax.set_yticklabels(self.yticklabels) if self.yticklabels is not None else ...
        ax.yaxis.set_minor_locator(AutoMinorLocator(self.n_minor))

        ax.set_xlim(self.xrange)
        ax.set_ylim(self.yrange)
        if self.hide_xticks:
            ax.tick_params(which="both", bottom=False)
        if self.hide_yticks:
            ax.tick_params(which="both", left=False)


class HeatFigSet(Struct):
    colorbar: bool = True
    """Whether plot colorbar"""

    cmap: Annotated[str, tyro.conf.arg(aliases=["-cm"])] = "cet_coolwarm"
    """Colormap"""

    norm: Annotated[
        Literal[
            "MyCustom",
            "Normal",
            "Logarithmic",
            "Centered",
            "SymmetricLogarithmic",
            "PowerLaw",
            "TwoSlopeNorm",
        ],
        tyro.conf.arg(name="normalization", aliases=["--norm"]),
    ] = "MyCustom"
    """Choose normalization method for colormap ,see  https://matplotlib.org/stable/users/explain/colors/colormapnorms.html"""

    power: float = 2.0
    """specific the power of PowerLaw Normalization"""

    symlogparm: tuple[float, float] = (0.1, 0.1)
    """specific linthresh and linscale for SymmetricLogarithmic Normalization"""

    cticks: list[float] | None = None
    """colorbar ticks"""

    cticklabels: list[str] | None = None
    """colorbar ticklabels for Figure"""

    vrange: tuple[float, float] | None = None

    vcenter: float = 0.0

    def __call__(self, mappable: "ScalarMappable"):
        arr = mappable.get_array()
        if arr is None:
            raise ValueError
        vmin, vmax = (arr.min()), arr.max()

        self._set_norm(mappable, vmin, vmax)
        if self.colorbar:
            self._set_cbar(mappable, vmin, vmax)

    def _set_norm(self, mappable: "ScalarMappable", vmin: float, vmax: float):
        from matplotlib import colors

        from wuxctools.utils.plot import MyCustomNormalize

        if self.norm == "Normal":
            norm = colors.Normalize(vmin, vmax)
        elif self.norm == "Logarithmic":
            if vmin < 0:
                raise ValueError(
                    "Logarithmic Normalization doesn't support negative values"
                )
            norm = colors.LogNorm(vmin, vmax)
        elif self.norm == "Centered":
            norm = colors.CenteredNorm(self.vcenter)
        elif self.norm == "SymmetricLogarithmic":
            norm = colors.SymLogNorm(
                linthresh=self.symlogparm[0],
                linscale=self.symlogparm[1],
                vmin=vmin,
                vmax=vmax,
            )
        elif self.norm == "TwoSlopeNorm":
            norm = colors.TwoSlopeNorm(vmin=vmin, vcenter=self.vcenter, vmax=vmax)
        elif self.norm == "PowerLaw":
            norm = colors.PowerNorm(self.power, vmin, vmax)
        else:
            norm = MyCustomNormalize(vmin, vmax, self.vcenter)

        mappable.set_norm(norm)

    def _set_cbar(self, mappable: "ScalarMappable", vmin: float, vmax: float):
        import colorcet  # pyright: ignore[reportMissingTypeStubs]

        _ = colorcet.__version__

        from wuxctools.utils import get_fig_ax

        mappable.set_cmap(self.cmap)
        fig, ax = get_fig_ax()
        cbar = fig.colorbar(mappable, ax=ax)
        if self.cticks is None:
            if vmin * vmax > 0:
                self.cticks = [
                    vmin,
                    (vmin + vmax) / 4,
                    (vmin + vmax) / 2,
                    (vmin + vmax) / 4 * 3,
                    vmax,
                ]
            else:
                self.cticks = [
                    vmin,
                    (vmin + self.vcenter) / 2,
                    self.vcenter,
                    (vmax + self.vcenter) / 2,
                    vmax,
                ]

        else:
            ...
        cbar.set_ticks(self.cticks)

        if self.cticklabels is None:
            self.cticklabels = [f"{i:.2f}" for i in self.cticks]
        cbar.set_ticklabels(self.cticklabels)



from pathlib import Path

import tyro
from msgspec import Struct, field

from wuxctools.plot import BaseFigSet


class Line(Struct, tag="Line"):
    """
    Plot Line figure
    """

    files: tyro.conf.Positional[list[Path]] = field(
        default_factory=lambda: [Path("data.txt")]
    )
    unpack: bool = True
    cols: tuple[int, int] = (0, 1)
    basefigset: BaseFigSet = field(default_factory=BaseFigSet)

    def plot_sigle_file(self, file: Path):
        import numpy as np

        from wuxctools.utils import get_fig_ax

        data = np.loadtxt(file, unpack=self.unpack)
        _, ax = get_fig_ax()
        ax.plot(data[self.cols[0]], data[self.cols[1]])
        self.basefigset()

    def __call__(self) -> None:
        import matplotlib.pyplot as plt

        from wuxctools.utils import plot_series, set_style

        set_style(self.basefigset.matplotlibrc)
        plot_series(self.files, self.basefigset.save, self.plot_sigle_file)

        if self.basefigset.show:
            plt.show()
